<?php

namespace Drupal\date_filter\Plugin\views\filter;

use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItem;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\views\FieldAPIHandlerTrait;
use Drupal\views\Plugin\views\display\DisplayPluginBase;
use Drupal\views\Plugin\views\filter\NumericFilter;
use Drupal\views\ViewExecutable;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * An improved Date/time views filter.
 *
 * Based on Drupal\datetime\Plugin\views\filter\Date, hopefully
 * it'll replace its origin one day.
 *
 * @ingroup views_filter_handlers
 *
 * @ViewsFilter("datetime_proper")
 */
class Date extends NumericFilter implements ContainerFactoryPluginInterface {

  use FieldAPIHandlerTrait;

  /**
   * The date format storage.
   */
  protected EntityStorageInterface $dateFormatStorage;

  /**
   * Some datetime fields don't use time at all.
   */
  protected bool $skipTime = FALSE;

  /**
   * Does this filter use date only? UI setting.
   */
  protected bool $skipTimeUi = TRUE;

  /**
   * Data type, date or timestamp.
   */
  protected string $dataType;

  /**
   * Possible value elements array, key -> label.
   *
   * @var \Drupal\Core\StringTranslation\TranslatableMarkup[]
   */
  protected array $valueElements;

  /**
   * Constructs a new Date handler.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityStorageInterface $date_format_storage
   *   The date format storage.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EntityStorageInterface $date_format_storage
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->dateFormatStorage = $date_format_storage;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager')->getStorage('date_format')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function init(ViewExecutable $view, DisplayPluginBase $display, ?array &$options = NULL) {
    parent::init($view, $display, $options);

    // We need definition for base fields as well and the trait method
    // relies on this property.
    if (!\array_key_exists('field_name', $this->definition)) {
      $this->definition['field_name'] = $this->definition['entity field'];
    }
    $definition = $this->getFieldStorageDefinition();

    $this->dataType = $definition->getPropertyDefinition('value')->getDataType();

    if (
      \class_exists(DateTimeItem::class) &&
      $definition->getSetting('datetime_type') === DateTimeItem::DATETIME_TYPE_DATE
    ) {
      $this->skipTime = TRUE;
    }
    if ($this->options['type'] === 'datetime' && !$this->skipTime) {
      $this->skipTimeUi = FALSE;
    }

    $this->valueElements = [
      'value' => NULL,
      'min' => $this->t('from'),
      'max' => $this->t('to'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function operators() {
    $operators = parent::operators();
    // @todo Will this be ever needed in a date filter?
    unset($operators['regular_expression']);

    return $operators;
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    $options = parent::defineOptions();

    $options['type'] = [
      'default' => 'date',
    ];
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    // This could be a checkbox really but we're staying with type
    // property and radios so we don't have to do anything with date
    // and datetime config schema.
    // @todo If this'll ever make its way to core, config update will be
    // needed to convert this to boolean.
    $form['type'] = [
      '#type' => 'radios',
      '#title' => $this->t('Filter type'),
      '#default_value' => $this->skipTimeUi ? 'date' : 'datetime',
      '#options' => [
        'date' => $this->t('Date'),
        'datetime' => $this->t('Date and time'),
      ],
    ];
    if ($this->skipTime) {
      $form['type']['#disabled'] = TRUE;
      $form['type']['#description'] = $this->t('This is a date-only field.');
    }

    parent::buildOptionsForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateOptionsForm(&$form, FormStateInterface $form_state) {
    parent::validateOptionsForm($form, $form_state);

    if (empty($this->options['expose']['required'])) {
      return;
    }

    $operators = $this->operators();
    $operator = $form_state->getValue(['options', 'operator']);
    if ($operators[$operator]['values'] == 1) {
      $validate_keys = ['value'];
    }
    elseif ($operators[$operator]['values'] == 2) {
      $validate_keys = ['min', 'max'];
    }

    // Validate that the time values convert to something usable.
    $value = $form_state->getValue(['options', 'value']);
    foreach ($validate_keys as $validate_key) {
      if (!\array_key_exists($validate_key, $form['value'])) {
        continue;
      }
      $convert = \strtotime($value[$validate_key]);
      if ($convert == -1 || $convert === FALSE) {
        $form_state->setError($form['value'][$validate_key], $this->t('Invalid date format.'));
      }
    }
  }

  /**
   * Add a type selector to the value form.
   */
  protected function valueForm(&$form, FormStateInterface $form_state): void {
    parent::valueForm($form, $form_state);

    if ($form_state->get('exposed') !== TRUE) {
      foreach (\array_keys($this->valueElements) as $value_key) {
        $form['value'][$value_key]['#description'] = $this->t('A date in any machine readable format (CCYY-MM-DD HH:MM:SS is preferred) or an offset from the current time such as "@example1" or "@example2".', [
          '@example1' => '+1 day',
          '@example2' => '-2 hours -30 minutes',
        ]);
      }
      return;
    }

    // Now we have end-user facing value form without Views UI modal
    // limitations and offset needs so datepickers are much nicer.
    // @todo Conditionally display datepickers if default value type
    // is not an offset in Views admin UI as well.
    // @todo Predefined offset options?
    $input = $form_state->getUserInput();
    $identifier = $this->options['expose']['identifier'];
    $formats = \explode('\T', $this->getStorageFormat($this->skipTimeUi));

    if (\array_key_exists('#type', $form['value'])) {
      $element = &$form;
      $nesting = FALSE;
    }
    else {
      $element = &$form['value'];
      $nesting = TRUE;
    }

    foreach ($this->valueElements as $value_key => $label) {
      if (!\array_key_exists($value_key, $element)) {
        continue;
      }
      $date = $this->getDate($this->value[$value_key]);
      $element[$value_key] = [
        '#type' => 'datetime',
        '#title' => $label,
        '#date_increment' => 1,
        '#date_date_format' => $formats[0],
        '#date_timezone' => $this->getInputTimezone(),
        '#date_time_element' => $this->skipTimeUi ? 'none' : 'time',
        '#date_time_format' => $this->skipTimeUi ? NULL : $formats[1],
        '#default_value' => $date,
      ];

      $this->modifyUserInput($input, $identifier, $nesting, $date, $value_key, $formats);

    }
    $form_state->setUserInput($input);
  }

  /**
   * Modify user input to ensure correct default values in the element.
   *
   * @param mixed[] $input
   *   The user input.
   * @param string $identifier
   *   The filter identifier.
   * @param bool $nesting
   *   Is the value nested?
   * @param \Drupal\Core\Datetime\DrupalDateTime|null $date
   *   The date to apply.
   * @param string $value_key
   *   The value key.
   * @param string[] $formats
   *   Date and time formats.
   */
  private function modifyUserInput(
    array &$input,
    string $identifier,
    bool $nesting,
    ?DrupalDateTime $date,
    string $value_key,
    array $formats
  ): void {
    if ($date === NULL) {
      return;
    }

    if ($nesting) {
      if (
        !\array_key_exists($identifier, $input) ||
        !\array_key_exists($value_key, $input[$identifier])
      ) {
        $input[$identifier][$value_key] = [
          'date' => $date->format($formats[0]),
          'time' => $this->skipTimeUi ? NULL : $date->format($formats[1]),
        ];
      }
    }
    else {
      if (
        !\array_key_exists($identifier, $input) ||
        !\is_array($input[$identifier])
      ) {
        $input[$identifier] = [
          'date' => $date->format($formats[0]),
          'time' => $this->skipTimeUi ? NULL : $date->format($formats[1]),
        ];
      }
    }
  }

  /**
   * Override parent method, which deals with dates as integers.
   */
  protected function opBetween($field): void {
    $value = [];
    foreach (['min', 'max'] as $value_key) {
      $value[$value_key] = $this->processValue($this->value[$value_key], $value_key);
    }

    if ($value['min'] !== NULL && $value['max'] !== NULL) {
      $operator = \strtoupper($this->operator);
      $this->query->addWhereExpression($this->options['group'], "$field $operator {$value['min']} AND {$value['max']}");
    }
    elseif ($value['min'] !== NULL) {
      $this->query->addWhereExpression($this->options['group'], "$field >= {$value['min']}");
    }
    elseif ($value['max'] !== NULL) {
      $this->query->addWhereExpression($this->options['group'], "$field <= {$value['max']}");
    }
  }

  /**
   * Override parent method, which deals with dates as integers.
   */
  protected function opSimple($field): void {
    // Special cases where we want to find content from a specific date.
    if ($this->operator === '=' && $this->skipTimeUi) {
      $this->value['min'] = $this->value['max'] = $this->value['value'];
      $this->operator = 'between';
      $this->opBetween($field);
      return;
    }
    if ($this->operator === '!=' && $this->skipTimeUi) {
      $this->value['min'] = $this->value['max'] = $this->value['value'];
      $this->operator = 'not between';
      $this->opBetween($field);
      return;
    }

    $value = $this->processValue($this->value['value']);
    if ($value === NULL) {
      return;
    }

    $this->query->addWhereExpression($this->options['group'], "$field $this->operator $value");
  }

  /**
   * Converts value to date if needed.
   *
   * @param \Drupal\Core\Datetime\DrupalDateTime|string|null $value
   *   The input value to process.
   * @param string $value_key
   *   Type of value: min, max or value.
   */
  protected function processValue($value, string $value_key = ''): ?string {
    if ($value === NULL) {
      return NULL;
    }
    if (\is_string($value) && $value === '') {
      return NULL;
    }
    if (\is_string($value)) {
      $value = $this->getDate($value);
    }
    if (!$value instanceof DrupalDateTime) {
      return NULL;
    }

    // Special case: date-only. We need to set times.
    if ($this->skipTimeUi) {
      $this->resetTimes($value, $value_key);
    }

    if ($this->dataType === 'timestamp') {
      return $value->format('U');
    }
    // The second known data type case here is 'datetime_iso8601'.
    // @todo Are there any other datetime storage types that'll not be covered?
    else {
      // Convert date to the storage format.
      $format = $this->getStorageFormat($this->skipTime);
      $date_value = $value->format($format, [
        'timezone' => $this->getStorageTimezone(),
      ]);

      $date_field = $this->query->getDateField(
        "'" . $date_value . "'",
        TRUE,
        FALSE
      );

      return $this->query->getDateFormat($date_field, $format, TRUE);
    }
  }

  /**
   * Reset times for proper date filtering.
   */
  protected function resetTimes(DrupalDateTime $date, string $value_key): void {
    if ($this->operator === '=' || $this->operator === '!=') {
      return;
    }

    if ($this->operator === '>' || $this->operator === '>=') {
      $date->setTime(0, 0, 0);
    }
    if ($this->operator === '<' || $this->operator === '<=') {
      $date->setTime(23, 59, 59);
    }
    if ($this->operator === 'between' || $this->operator === 'not between') {
      if ($value_key === 'min') {
        $date->setTime(0, 0, 0);
      }
      elseif ($value_key === 'max') {
        $date->setTime(23, 59, 59);
      }
    }

  }

  /**
   * Get date from offset to use in computations.
   *
   * @param string $date_string
   *   The date offset.
   *
   * @return \Drupal\Core\Datetime\DrupalDateTime|null
   *   DrupalDateTime object representing user input or NULL.
   */
  protected function getDate(string $date_string): ?DrupalDateTime {
    if ($date_string === '') {
      return NULL;
    }
    $date_string = \trim($date_string);
    $date = new DrupalDateTime($date_string, $this->getInputTimezone());

    return $date;
  }

  /**
   * Get datetime storage format.
   */
  protected function getStorageFormat(bool $no_time = FALSE): string {
    if ($no_time) {
      return \interface_exists(DateTimeItemInterface::class) ? DateTimeItemInterface::DATE_STORAGE_FORMAT : 'Y-m-d';
    }
    return \interface_exists(DateTimeItemInterface::class) ? DateTimeItemInterface::DATETIME_STORAGE_FORMAT : 'Y-m-d\TH:i:s';
  }

  /**
   * Get datetime storage timezone.
   */
  protected function getStorageTimezone() {
    return \interface_exists(DateTimeItemInterface::class) ? DateTimeItemInterface::STORAGE_TIMEZONE : 'UTC';
  }

  /**
   * Gets time zone depending on date storage type.
   *
   * The general rule here is that date-only datetime fields
   * need to ignore the user time zone and treat the input as
   * it was in the starage time zone. Everything else respects
   * time zones.
   */
  protected function getInputTimezone(): string {
    if ($this->dataType === 'timestamp') {
      return \date_default_timezone_get();
    }

    if (!$this->skipTime) {
      return \date_default_timezone_get();
    }

    return \interface_exists(DateTimeItemInterface::class) ? DateTimeItemInterface::STORAGE_TIMEZONE : 'UTC';
  }

}
